import numpy as np
import sfs_da.util as util
import sfs_da.pivot as pivot 
import sfs_da.conditioning as conditioning

def divide_and_conquer_Lasso(X, a, b, ns, nt, Lambda, zmin, zmax):
    list_intervals = []
    list_M = []

    zuv = zmin

    while zuv <= zmax:
        Y_zuv = a + b * zuv

        # Solve Linear Program to obtain Transportation Map for OT
        H, h, Theta, c_, cost = util.construct(X, Y_zuv, ns, nt)
        Bu, Bcu, Tu = util.LinearProgram(cost, H, h)
        Tu = Tu.reshape(ns, nt)
        Omega_u = np.hstack((np.zeros((ns + nt, ns)), np.vstack((ns * Tu, np.identity(nt)))))

        interval_u = conditioning.conditioning_Zu(a, b, c_, Theta, Bu, Bcu, H)

        # Select the interval containing the data point that we are currently considering.
        for i in interval_u:
            if i[0] <= zuv <= i[1]:
                interval_u = [i]
                break

        Xtilde_u = Omega_u.dot(X)

        while zuv <= interval_u[0][1]:
            Y_zuv = a + b * zuv
            Ytilde_u_zuv = Omega_u.dot(Y_zuv)

            M_v, Mc_v, sM_v = util.solve_Lasso(Xtilde_u, Ytilde_u_zuv, Lambda)

            interval_v = conditioning.conditioning_Zv_Lasso(Xtilde_u, a, b, Lambda, Omega_u, M_v, Mc_v, sM_v)
            
            interval_uv = util.intersect(interval_u, interval_v)

            list_intervals += interval_uv
            list_M += [M_v]
            zuv = interval_uv[0][1] + 1e-4

    return list_intervals, list_M

def SFS_DA_Lasso(Xs, Ys, Xt, Yt, Lambda, Sigma_s, Sigma_t, zmin=-20, zmax=20):
    ns = Ys.shape[0]
    nt = Yt.shape[0]
    
    X = np.vstack((Xs, Xt))
    Y = np.vstack((Ys, Yt))
    Sigma = np.vstack((np.hstack((Sigma_s, np.zeros((ns, nt)))), np.hstack((np.zeros((nt, ns)), Sigma_t))))

    M_obs = util.Lasso_after_DA(X, Y, ns, nt, Lambda)[0]
    k = M_obs.shape[0]

    if k == 0:
        return None

    Xt_M = Xt[:, M_obs]

    list_p_sel = []

    for j in range(k):
        ej = np.zeros((k, 1))
        ej[j][0] = 1

        etaj = np.vstack((np.zeros((ns, 1)), Xt_M.dot(np.linalg.inv(Xt_M.T.dot(Xt_M))).dot(ej)))
        
        b = Sigma.dot(etaj).dot(np.linalg.inv(etaj.T.dot(Sigma).dot(etaj)))
        a = (np.identity(ns + nt) - b.dot(etaj.T)).dot(Y)

        list_intervals, list_M = divide_and_conquer_Lasso(X, a, b, ns, nt, Lambda, zmin, zmax)

        Z = []
        for i in range(len(list_intervals)):
            if np.array_equal(list_M[i], M_obs):
                Z.append(list_intervals[i])

        # Compute p-value
        etajTY = etaj.T.dot(Y)[0][0]
        etajT_Sigma_etaj = etaj.T.dot(Sigma).dot(etaj)[0][0]

        pj_sel = pivot.compute_TN_p_value(Z, etajTY, etajT_Sigma_etaj, 0)

        list_p_sel.append([pj_sel, M_obs[j]])

    return list_p_sel

def SFS_DA_Lasso_rand_feature(Xs, Ys, Xt, Yt, Lambda, Sigma_s, Sigma_t, zmin=-20, zmax=20):
    ns = Ys.shape[0]
    nt = Yt.shape[0]
    
    X = np.vstack((Xs, Xt))
    Y = np.vstack((Ys, Yt))
    Sigma = np.vstack((np.hstack((Sigma_s, np.zeros((ns, nt)))), np.hstack((np.zeros((nt, ns)), Sigma_t))))

    M_obs = util.Lasso_after_DA(X, Y, ns, nt, Lambda)[0]
    k = M_obs.shape[0]

    if k == 0:
        return None

    Xt_M = Xt[:, M_obs]

    j = np.random.randint(k)
    ej = np.zeros((k, 1))
    ej[j][0] = 1

    etaj = np.vstack((np.zeros((ns, 1)), Xt_M.dot(np.linalg.inv(Xt_M.T.dot(Xt_M))).dot(ej)))
    
    b = Sigma.dot(etaj).dot(np.linalg.inv(etaj.T.dot(Sigma).dot(etaj)))
    a = (np.identity(ns + nt) - b.dot(etaj.T)).dot(Y)

    list_intervals, list_M = divide_and_conquer_Lasso(X, a, b, ns, nt, Lambda, zmin, zmax)

    Z = []
    for i in range(len(list_intervals)):
        if np.array_equal(list_M[i], M_obs):
            Z.append(list_intervals[i])

    # Compute p-value
    etajTY = etaj.T.dot(Y)[0][0]
    etajT_Sigma_etaj = etaj.T.dot(Sigma).dot(etaj)[0][0]

    pj_sel = pivot.compute_TN_p_value(Z, etajTY, etajT_Sigma_etaj, 0)

    return pj_sel

def divide_and_conquer_ElasticNet(X, a, b, ns, nt, Lambda, Gamma, zmin, zmax):
    list_intervals = []
    list_M = []

    zuv = zmin

    while zuv <= zmax:
        Y_zuv = a + b * zuv

        # OT from source to target
        H, h, Theta, c_, cost = util.construct(X, Y_zuv, ns, nt)
        Bu, Bcu, Tu = util.LinearProgram(cost, H, h)
        Tu = Tu.reshape(ns, nt)
        Omega_u = np.hstack((np.zeros((ns + nt, ns)), np.vstack((ns * Tu, np.identity(nt)))))

        interval_u = conditioning.conditioning_Zu(a, b, c_, Theta, Bu, Bcu, H)

        # Select the interval containing the data point that we are currently considering.
        for i in interval_u:
            if i[0] <= zuv <= i[1]:
                interval_u = [i]
                break

        Xtilde_u = Omega_u.dot(X)

        while zuv <= interval_u[0][1]:
            Y_zuv = a + b * zuv
            Ytilde_u_zuv = Omega_u.dot(Y_zuv)

            M_v, Mc_v, sM_v = util.solve_ElasticNet(Xtilde_u, Ytilde_u_zuv, Lambda, Gamma)

            interval_v = conditioning.conditioning_Zv_ElasticNet(Xtilde_u, a, b, Lambda, Gamma, Omega_u, M_v, Mc_v, sM_v)
            
            interval_uv = util.intersect(interval_u, interval_v)

            list_intervals += interval_uv
            list_M += [M_v]
            zuv = interval_uv[0][1] + 1e-4

    return list_intervals, list_M

def SFS_DA_ElasticNet(Xs, Ys, Xt, Yt, Lambda, Gamma, Sigma_s, Sigma_t, zmin=-20, zmax=20):
    ns = Ys.shape[0]
    nt = Yt.shape[0]
    
    X = np.vstack((Xs, Xt))
    Y = np.vstack((Ys, Yt))
    Sigma = np.vstack((np.hstack((Sigma_s, np.zeros((ns, nt)))), np.hstack((np.zeros((nt, ns)), Sigma_t))))

    M_obs = util.EN_after_DA(X, Y, ns, nt, Lambda, Gamma)[0]
    k = M_obs.shape[0]

    if k == 0:
        return None

    Xt_M = Xt[:, M_obs]

    list_p_sel = []

    for j in range(k):
        ej = np.zeros((k, 1))
        ej[j][0] = 1

        etaj = np.vstack((np.zeros((ns, 1)), Xt_M.dot(np.linalg.inv(Xt_M.T.dot(Xt_M))).dot(ej)))
        
        b = Sigma.dot(etaj).dot(np.linalg.inv(etaj.T.dot(Sigma).dot(etaj)))
        a = (np.identity(ns + nt) - b.dot(etaj.T)).dot(Y)

        list_intervals, list_M = divide_and_conquer_ElasticNet(X, a, b, ns, nt, Lambda, Gamma, zmin, zmax)

        Z = []
        for i in range(len(list_intervals)):
            if np.array_equal(list_M[i], M_obs):
                Z.append(list_intervals[i])

        # Compute p-value
        etajTY = etaj.T.dot(Y)[0][0]
        etajT_Sigma_etaj = etaj.T.dot(Sigma).dot(etaj)[0][0]

        pj_sel = pivot.compute_TN_p_value(Z, etajTY, etajT_Sigma_etaj, 0)

        list_p_sel.append([pj_sel, M_obs[j]])

    return list_p_sel

def SFS_DA_ElasticNet_rand_feature(Xs, Ys, Xt, Yt, Lambda, Gamma, Sigma_s, Sigma_t, zmin=-20, zmax=20):
    ns = Ys.shape[0]
    nt = Yt.shape[0]
    
    X = np.vstack((Xs, Xt))
    Y = np.vstack((Ys, Yt))
    Sigma = np.vstack((np.hstack((Sigma_s, np.zeros((ns, nt)))), np.hstack((np.zeros((nt, ns)), Sigma_t))))

    M_obs = util.EN_after_DA(X, Y, ns, nt, Lambda, Gamma)[0]
    k = M_obs.shape[0]

    if k == 0:
        return None

    Xt_M = Xt[:, M_obs]

    j = np.random.randint(k)

    ej = np.zeros((k, 1))
    ej[j][0] = 1

    etaj = np.vstack((np.zeros((ns, 1)), Xt_M.dot(np.linalg.inv(Xt_M.T.dot(Xt_M))).dot(ej)))
    
    b = Sigma.dot(etaj).dot(np.linalg.inv(etaj.T.dot(Sigma).dot(etaj)))
    a = (np.identity(ns + nt) - b.dot(etaj.T)).dot(Y)

    list_intervals, list_M = divide_and_conquer_ElasticNet(X, a, b, ns, nt, Lambda, Gamma, zmin, zmax)

    Z = []
    for i in range(len(list_intervals)):
        if np.array_equal(list_M[i], M_obs):
            Z.append(list_intervals[i])

    # Compute p-value
    etajTY = etaj.T.dot(Y)[0][0]
    etajT_Sigma_etaj = etaj.T.dot(Sigma).dot(etaj)[0][0]

    pj_sel = pivot.compute_TN_p_value(Z, etajTY, etajT_Sigma_etaj, 0)

    return pj_sel


import numpy as np
import sfs_da.util as util

def solve_linear_unequalities(left, right):
    # Solve unequality has the form of ax >= b
    
    Z_minus = np.NINF
    Z_plus = np.inf

    for i in range(left.shape[0]):
        l = left[i][0]
        r = right[i][0]

        if - 1e-10 <= l <= 1e-10:
            l = 0.0

        if l == 0.0:
            continue

        else:
            temp = r / l

            if l > 0.0:
                Z_minus = max(temp, Z_minus)
            else:
                Z_plus = min(temp, Z_plus)

    return [[Z_minus, Z_plus]]

def solve_quadratic_inequality(a, b, c):
    # Solve quadratic inequality has the form of ax^2 + bx + c <= 0
    if -1e-7 <= a <= 1e-7:
        a = 0.0

    if -1e-7 <= b <= 1e-7:
        b = 0.0

    if -1e-7 <= c <= 1e-7:
        c = 0.0

    if a == 0.0:
        if b != 0.0:
            root = np.round(-c/b, 12)

            if b > 0.0:
                return [[np.NINF, root]]
            else:
                return [[root, np.inf]]
        
        else:
            if c > 0.0:
                # print("Error no roots: c > 0")
                return None
            else:
                return [[np.NINF, np.inf]]
            
    delta = b**2 - 4 * a * c

    if delta < 0.0:
        if a > 0.0:
            # print("Error no roots: a > 0")
            return None
        else: 
            return [[np.NINF, np.inf]]
    else:
        sqrt_delta = np.sqrt(delta)

        if b > 0:
            root1 = np.round((-b-sqrt_delta)/(2*a), 12)
        else:
            root1 = np.round((-b+sqrt_delta)/(2*a), 12)

        root2 = np.round(c / (a * root1), 12)

        roots = np.sort([root1, root2])
        
        if a > 0:
            return [roots]
        else:
            return [[np.NINF, roots[0]], [roots[1], np.inf]]
            
def conditioning_Zu(a, b, c_, Theta, B, Bc, H):
    Theta_a = Theta.dot(a)
    Theta_b = Theta.dot(b)

    p_tilde = c_ + Theta_a * Theta_a
    q_tilde = Theta_a * Theta_b + Theta_b * Theta_a
    r_tilde = Theta_b * Theta_b

    H_B_invH_Bc = np.linalg.inv(H[:, B]).dot(H[:, Bc])

    p = (p_tilde[Bc, :].T - p_tilde[B, :].T.dot(H_B_invH_Bc)).T
    q = (q_tilde[Bc, :].T - q_tilde[B, :].T.dot(H_B_invH_Bc)).T
    r = (r_tilde[Bc, :].T - r_tilde[B, :].T.dot(H_B_invH_Bc)).T

    flag = False
    list_intervals = []

    for i in range(p.shape[0]):
        fa = - r[i][0]
        sa = - q[i][0]
        ta = - p[i][0]

        temp = solve_quadratic_inequality(fa, sa, ta)
        
        if flag == False:
            flag = True
            list_intervals = temp
        else:
            list_intervals = util.intersect(list_intervals, temp)

    return list_intervals

def conditioning_Zv_Lasso(Xtilde, a, b, Lambda, Omega, M, Mc, sM):
    XtildeM = Xtilde[:, M]
    XtildeMc = Xtilde[:, Mc]

    XtildeMT_XtildeM_inv = np.linalg.inv(XtildeM.T.dot(XtildeM))
    XtildeM_plus = XtildeMT_XtildeM_inv.dot(XtildeM.T)
    XtildeMT_plus = XtildeM.dot(XtildeMT_XtildeM_inv)

    # Construct Atilde0 and Btilde0
    Atilde0 = - sM * (XtildeM_plus).dot(Omega) 
    Btilde0 = - Lambda * sM * (XtildeMT_XtildeM_inv.dot(sM))

    # Construct Atilde1 and Btilde1
    Atilde10 = XtildeMc.T.dot(np.identity(Xtilde.shape[0]) - XtildeM.dot(XtildeM_plus)).dot(Omega)
    Atilde11 = - Atilde10

    tmp = XtildeMc.T.dot(XtildeMT_plus).dot(sM)
    Btilde10 = np.ones((tmp.shape[0], 1)) - tmp
    Btilde11 = np.ones((tmp.shape[0], 1)) + tmp

    Atilde1 = np.vstack((Atilde10, Atilde11)) / Lambda
    Btilde1 = np.vstack((Btilde10, Btilde11))

    Atilde = np.vstack((Atilde0, Atilde1))
    Btilde = np.vstack((Btilde0, Btilde1))

    A = Atilde.dot(b)
    B = Btilde - Atilde.dot(a)
    return solve_linear_unequalities(-A, -B)

def conditioning_Zv_ElasticNet(Xtilde, a, b, Lambda, Gamma, Omega, M, Mc, sM):
    XtildeM = Xtilde[:, M]
    XtildeMc = Xtilde[:, Mc]

    XtildeMT_XtildeM_inv = np.linalg.inv(XtildeM.T.dot(XtildeM) + Gamma * np.identity(XtildeM.shape[1]))
    XtildeM_star = XtildeMT_XtildeM_inv.dot(XtildeM.T)
    XtildeMT_star = XtildeM.dot(XtildeMT_XtildeM_inv)

    # Construct Atilde0 and Btilde0
    Atilde0 = - sM * (XtildeM_star).dot(Omega) 
    Btilde0 = - Lambda * sM * (XtildeMT_XtildeM_inv.dot(sM))

    # Construct Atilde1 and Btilde1
    Atilde10 = XtildeMc.T.dot(np.identity(Xtilde.shape[0]) - XtildeM.dot(XtildeM_star)).dot(Omega)
    Atilde11 = - Atilde10

    tmp = XtildeMc.T.dot(XtildeMT_star).dot(sM)
    Btilde10 = np.ones((tmp.shape[0], 1)) - tmp
    Btilde11 = np.ones((tmp.shape[0], 1)) + tmp

    Atilde1 = np.vstack((Atilde10, Atilde11)) / Lambda
    Btilde1 = np.vstack((Btilde10, Btilde11))

    Atilde = np.vstack((Atilde0, Atilde1))
    Btilde = np.vstack((Btilde0, Btilde1))

    A = Atilde.dot(b)
    B = Btilde - Atilde.dot(a)
    return solve_linear_unequalities(-A, -B)


import numpy as np

def gen_data(n, m, p, true_beta_s, true_beta_t):
    Xs = np.random.normal(loc = 0, scale = 1, size = (n, p))
    Xt = np.random.normal(loc = 0, scale = 1, size = (m, p))

    mu_s = Xs.dot(true_beta_s)
    mu_t = Xt.dot(true_beta_t)

    Ys = mu_s + np.random.normal(loc = 0, scale = 1, size = (n, 1))
    Yt = mu_t + np.random.normal(loc = 0, scale = 1, size = (m, 1))

    Sigma_s = np.identity(n)
    Sigma_t = np.identity(m)

    return Xs, Xt, Ys, Yt, Sigma_s, Sigma_t


import numpy as np
from scipy.optimize import linprog
from sklearn.linear_model import Lasso
from sklearn.linear_model import ElasticNet

def construct(X, Y, ns, nt):
    Mr = np.zeros((ns, ns * nt))
    
    for i in range(ns):
        Mr[i : i + 1, i * nt : (i + 1) * nt] = np.ones((1, nt))
        
    Mc = np.identity(nt)
    for _ in range(ns - 1):
        Mc = np.hstack((Mc, np.identity(nt)))

    S = np.vstack((Mr, Mc))

    h = np.vstack((np.ones((ns, 1)) / ns, np.ones((nt, 1)) / nt))

    # Delete last row
    S = S[0 : (ns + nt - 1), :]
    h = h[0 : (ns + nt - 1), :]

    Omega = np.hstack((np.kron(np.identity(ns), np.ones((nt, 1))), np.kron(- np.ones((ns, 1)), np.identity(nt))))
    
    c_ = np.zeros((ns * nt, 1))

    for i in range(X.shape[1]):
        c_ += (Omega.dot(X[:, [i]])) * (Omega.dot(X[:, [i]]))

    cost = c_ + (Omega.dot(Y)) * (Omega.dot(Y)) 

    return S, h, Omega, c_, cost

def LinearProgram(cost, S, h):
    n = cost.shape[0]
    res = linprog(cost, A_ub = - np.identity(n), b_ub = np.zeros((n, 1)), A_eq = S, b_eq = h, method = 'simplex', options = {'maxiter': 20000})

    B = res.basis
    Bc = np.arange(n)

    for i in B:
        Bc = np.delete(Bc, np.where(Bc == i)[0][0])

    return B, Bc, res.x

def solve_Lasso(X, Y, Lambda):
    p = X.shape[1]

    lasso = Lasso(alpha = Lambda / X.shape[0], fit_intercept=False, tol = 1e-10, max_iter=100000000)
    lasso.fit(X, Y)

    beta_hat = lasso.coef_.reshape(p, 1)
    M = np.nonzero(beta_hat)[0]
    Mc = np.nonzero(beta_hat == 0)[0]
    sM = np.sign(beta_hat[M])
    return M, Mc, sM

def solve_ElasticNet(X, Y, Lambda, Gamma):
    n = X.shape[0]
    p = X.shape[1]

    elasticnet = ElasticNet(alpha=(Lambda+Gamma)/n, l1_ratio=Lambda/(Lambda+Gamma), fit_intercept=False, tol = 1e-10, max_iter=10000000)
    elasticnet.fit(X, Y)

    beta_hat = elasticnet.coef_.reshape(p, 1)
    M = np.nonzero(beta_hat)[0]
    Mc = np.nonzero(beta_hat == 0)[0]
    sM = np.sign(beta_hat[M])
    return M, Mc, sM

def checkKKT_Lasso(X, Y, M, beta_hat, Lambda):
    XM = X[:, M]
    sM = np.sign(beta_hat)[M, :]
    beta_hat_M = beta_hat[M, :]
    print('-----------------Check KKT Lasso-----------------')
    print(np.round((XM.T.dot(XM))).dot(beta_hat_M) - XM.T.dot(Y) + Lambda * sM)
    print('-----------------Check KKT Lasso-----------------')

def checkKKT_EN(X, Y, M, beta_hat, Lambda, Gamma):
    XM = X[:, M]
    sM = np.sign(beta_hat)[M, :]
    beta_hat_M = beta_hat[M, :]
    print('-----------------Check KKT Elastic Net-----------------')
    print(np.round((XM.T.dot(XM) + Gamma * np.identity(M.shape[0])).dot(beta_hat_M) - XM.T.dot(Y) + Lambda * sM))
    print('-----------------Check KKT Elastic Net-----------------')

def intersect(list_intervals_1, list_intervals_2):
    results = []
    
    if len(list_intervals_1) == 0 or len(list_intervals_2) == 0:
        return results

    for interval_1 in list_intervals_1:
        for interval_2 in list_intervals_2:
            if interval_1[1] < interval_2[0]:
                break
            if interval_1[0] > interval_2[1]:
                continue
            results.append([max(interval_1[0], interval_2[0]), min(interval_1[1], interval_2[1])])

    return results

def Lasso_after_DA(X, Y, ns, nt, Lambda):
    p = X.shape[1]

    H, h, cost = [construct(X, Y, ns, nt)[i] for i in [0, 1, 4]]
    
    T = LinearProgram(cost, H, h)[2]
    T = T.reshape(ns, nt)

    Omega = np.hstack((np.zeros((ns + nt, ns)), np.vstack((ns * T, np.identity(nt)))))

    X_tilde = Omega.dot(X)
    Y_tilde = Omega.dot(Y)
    
    return solve_Lasso(X_tilde, Y_tilde, Lambda)
    
def EN_after_DA(X, Y, ns, nt, Lambda, Gamma):
    p = X.shape[1]

    S, h, cost = [construct(X, Y, ns, nt)[i] for i in [0, 1, 4]]
    
    T = LinearProgram(cost, S, h)[2]
    T = T.reshape(ns, nt)

    Omega = np.hstack((np.zeros((ns + nt, ns)), np.vstack((ns * T, np.identity(nt)))))

    X_tilde = Omega.dot(X)
    Y_tilde = Omega.dot(Y)
    
    return solve_ElasticNet(X_tilde, Y_tilde, Lambda, Gamma)


import numpy as np
from mpmath import mp
mp.dps = 500

def compute_TN_p_value(list_intervals, Test_Statistic, Var, mu_0):
    if len(list_intervals) == 0:
        print('Error no interval')
        return None

    tn_sigma = np.sqrt(Var)

    list_tn_cdf = []
    for interval in list_intervals:
        temp = mp.ncdf((interval[1] - mu_0) / tn_sigma) - mp.ncdf((interval[0] - mu_0) / tn_sigma)
        list_tn_cdf.append(temp)

    numerator = 0
    for i in range(len(list_intervals)):
        interval = list_intervals[i]

        if Test_Statistic > interval[1]:
            numerator += list_tn_cdf[i]
        else:
            numerator += mp.ncdf((Test_Statistic - mu_0) / tn_sigma) - mp.ncdf((interval[0] - mu_0) / tn_sigma)
            break

    denominator = sum(list_tn_cdf)

    if denominator == 0.0:
        print('Numerical error')
        return None

    cdf = float(numerator / denominator)

    p_value = 2 * min(cdf, 1 - cdf)
    return p_value


from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='sfs_da', 
    version='1.0.4',
    packages=find_packages(),
    install_requires=[
        'numpy',
        'mpmath',
        'scipy',
        'scikit-learn'
    ],
    description='A package for Statistical Inference for Feature Selection after OT-based Domain Adaptation',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Nguyen Thang Loi',
    author_email='23520872@gm.uit.edu.vn',
    license='MIT',
    url='https://github.com/NT-Loi/SFS_DA.git'
)